// (C) Copyright 2003: Reece H. Dunn

#ifndef BOOST__IOFM__DETAIL__TYPE_DEDUCER__HPP
#define BOOST__IOFM__DETAIL__TYPE_DEDUCER__HPP
   // all supported complex types:

#  include <utility>                             // std::pair
#  include <complex>                             // std::complex
#  include <boost/compressed_pair.hpp>           // boost::compressed_pair
#  include <boost/rational.hpp>                  // boost::rational

#  if !defined(BOOST_IOFM__NO_LIB_INTERVAL)
#     include <boost/numeric/interval/interval.hpp> // boost::numeric::interval
#  endif
#  if !defined(BOOST_IOFM__NO_LIB_QUATERNION)
#     include <boost/math/quaternion.hpp>           // boost::math::quaternion
#  endif
#  if !defined(BOOST_IOFM__NO_LIB_OCTONION)
#     include <boost/math/octonion.hpp>             // boost::math::octonion
#  endif

#  include <vector>    // std::vector
#  include <list>      // std::list
#  include <deque>     // std::deque
#  include <map>       // std::map
#  include <set>       // std::set

   namespace boost { namespace io
   {
      // type information - for mapping an outputter to the type

      struct io_type
      {
         enum
         {
            basic     = 0, // basic_output
            pair      = 1, // pair_output
            container = 2, // container_output
            array     = 3, // array_output
            nary      = 4  // static_nary_output
         };
      };

      // type deduction - basic

      template< typename T >
      struct type_deducer
      {
         enum{ type = boost::io::io_type::basic };

         typedef T                               object_type;
      };

      // type deduction - pair

      template< typename T1, typename T2 >
      struct type_deducer< std::pair< T1, T2 > >
      {
         enum{ type = boost::io::io_type::pair };

         typedef T1                              base1_type;
         typedef T2                              base2_type;

         typedef std::pair< T1, T2 >             object_type;
      };

      template< typename T1, typename T2 >
      struct type_deducer< boost::compressed_pair< T1, T2 > >
      {
         enum{ type = boost::io::io_type::pair };

         typedef T1                              base1_type;
         typedef T2                              base2_type;

         typedef boost::compressed_pair< T1, T2 >
                                                 object_type;
         };

      template< typename T >
      struct type_deducer< std::complex< T > >
      {
         enum{ type = boost::io::io_type::pair };

         typedef T                               base1_type;
         typedef T                               base2_type;

         typedef std::complex< T >               object_type;
      };

      template< typename T >
      struct type_deducer< boost::rational< T > >
      {
         enum{ type = boost::io::io_type::pair };

         typedef T                               base1_type;
         typedef T                               base2_type;

         typedef boost::rational< T >            object_type;
       };

#     if !defined(BOOST_IOFM__NO_LIB_INTERVAL)
         template< typename T, class Traits >
         struct type_deducer< boost::numeric::interval< T, Traits > >
         {
            enum{ type = boost::io::io_type::pair };

            typedef T                            base1_type;
            typedef T                            base2_type;

            typedef boost::numeric::interval< T, Traits >
                                                 object_type;
         };
#     endif

      // type deduction - container

      template< typename T, class Allocator >
      struct type_deducer< std::vector< T, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::vector< T, Allocator >     object_type;

         typedef typename object_type::value_type
                                                 base_type;
      };

      template< typename T, class Allocator >
      struct type_deducer< std::list< T, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::list< T, Allocator >       object_type;

         typedef typename object_type::value_type
                                                 base_type;
         };

      template< typename T, class Allocator >
      struct type_deducer< std::deque< T, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::deque< T, Allocator >      object_type;

         typedef typename object_type::value_type
                                                 base_type;
      };

      template< typename KeyT, typename T, class Cmp, class Allocator >
      struct type_deducer< std::map< KeyT, T, Cmp, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::map< KeyT, T, Cmp, Allocator >
                                                 object_type;

         typedef typename object_type::value_type
                                                 base_type;
      };

      template< typename KeyT, typename T, class Cmp, class Allocator >
      struct type_deducer< std::multimap< KeyT, T, Cmp, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::multimap< KeyT, T, Cmp, Allocator >
                                                 object_type;

         typedef typename object_type::value_type
                                                 base_type;
      };

      template< typename KeyT, class Cmp, class Allocator >
      struct type_deducer< std::set< KeyT, Cmp, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::set< KeyT, Cmp, Allocator >
                                                 object_type;

         typedef typename object_type::value_type
                                                 base_type;
      };

      template< typename KeyT, class Cmp, class Allocator >
      struct type_deducer< std::multiset< KeyT, Cmp, Allocator > >
      {
         enum{ type = boost::io::io_type::container };

         typedef std::multiset< KeyT, Cmp, Allocator >
                                                 object_type;

         typedef typename object_type::value_type
                                                 base_type;
      };

      // type deduction - array

      template< typename T, size_t N >
      struct type_deducer< T[ N ] >
      {
         enum{ type = boost::io::io_type::array };

         typedef T                               base_type;
         static const size_t                     length = N;

         typedef T                               object_type[ N ];
      };

      template< typename T, size_t N >
      struct type_deducer< const T[ N ] >
      {
         enum{ type = boost::io::io_type::array };

         typedef T                               base_type;
         static const size_t                     length = N;

         typedef const T                         object_type[ N ];
      };

      // type deduction - static n-ary

#     if !defined(BOOST_IOFM__NO_LIB_QUATERNION)
         template< typename T >
         struct type_deducer< boost::math::quaternion< T > >
         {
            enum{ type = boost::io::io_type::nary };

            typedef T                               base_type;
            typedef boost::math::quaternion< T >    object_type;
         };
#     endif

#     if !defined(BOOST_IOFM__NO_LIB_OCTONION)
         template< typename T >
         struct type_deducer< boost::math::octonion< T > >
         {
            enum{ type = boost::io::io_type::nary };

            typedef T                               base_type;
            typedef boost::math::octonion< T >      object_type;
         };
#     endif
   }}
#endif
